"""Tests for delay properties in the Opentrons protocol API."""

from pydantic import ValidationError
import pytest
from typing import Any, Union
from hypothesis import given, strategies as st, settings

from opentrons.protocol_api._liquid_properties import _build_touch_tip_properties
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    TouchTipProperties,
    LiquidClassTouchTipParams,
)

from . import (
    boolean_looking_values,
    invalid_values,
    positive_non_zero_floats_and_ints,
    reasonable_numbers,
    negative_or_zero_floats_and_ints,
)


def test_touch_tip_properties_enable_and_disable() -> None:
    """Test enabling and disabling TouchTipProperties."""
    tp = _build_touch_tip_properties(
        TouchTipProperties(
            enable=False,
            params=LiquidClassTouchTipParams(zOffset=1, mmFromEdge=1, speed=100),
        )
    )
    tp.enabled = True
    assert tp.enabled is True
    tp.enabled = False
    assert tp.enabled is False


def test_touch_tip_properties_none_instantiation_combos() -> None:
    """Test handling of None combinations in TouchTipProperties instantiation."""
    with pytest.raises(ValidationError):
        _build_touch_tip_properties(
            TouchTipProperties(enable=True, params=LiquidClassTouchTipParams(zOffset=None, mmFromEdge=None, speed=None))  # type: ignore
        )
    with pytest.raises(ValidationError):
        _build_touch_tip_properties(
            TouchTipProperties(enable=None, params=LiquidClassTouchTipParams(zOffset=None, mmFromEdge=1, speed=1))  # type: ignore
        )
    with pytest.raises(ValidationError):
        _build_touch_tip_properties(
            TouchTipProperties(enable=True, params=LiquidClassTouchTipParams(zOffset=1, mmFromEdge=None, speed=1))  # type: ignore
        )
    with pytest.raises(ValidationError):
        _build_touch_tip_properties(
            TouchTipProperties(enable=True, params=LiquidClassTouchTipParams(zOffset=1, mmFromEdge=1, speed=None))  # type: ignore
        )


@given(bad_value=st.one_of(invalid_values, boolean_looking_values))
@settings(deadline=None, max_examples=50)
def test_touch_tip_properties_enabled_bad_values(bad_value: Any) -> None:
    """Test bad values for TouchTipProperties.enabled."""
    with pytest.raises(ValidationError):
        _build_touch_tip_properties(
            TouchTipProperties(
                enable=bad_value,
                params=LiquidClassTouchTipParams(zOffset=1, mmFromEdge=1, speed=1),
            )
        )
    tp = _build_touch_tip_properties(
        TouchTipProperties(
            enable=False,
            params=LiquidClassTouchTipParams(zOffset=1, mmFromEdge=1, speed=1),
        )
    )
    with pytest.raises(ValueError):
        tp.enabled = bad_value


@given(good_value=reasonable_numbers)
@settings(deadline=None, max_examples=50)
def test_touch_tip_properties_z_offset(good_value: Union[int, float]) -> None:
    """Test valid z_offset."""
    _build_touch_tip_properties(
        TouchTipProperties(
            enable=True,
            params=LiquidClassTouchTipParams(
                zOffset=good_value, mmFromEdge=1, speed=10
            ),
        )
    )
    tp = _build_touch_tip_properties(
        TouchTipProperties(
            enable=False,
            params=LiquidClassTouchTipParams(zOffset=0, mmFromEdge=1, speed=10),
        )
    )
    tp.z_offset = good_value
    assert tp.z_offset == float(good_value)


@given(bad_value=invalid_values)
@settings(deadline=None, max_examples=50)
def test_touch_tip_properties_z_offset_bad_values(bad_value: Any) -> None:
    """Test invalid z_offset values."""
    with pytest.raises(ValidationError):
        _build_touch_tip_properties(
            TouchTipProperties(
                enable=True,
                params=LiquidClassTouchTipParams(
                    zOffset=bad_value, mmFromEdge=1, speed=10
                ),
            )
        )
    tp = _build_touch_tip_properties(
        TouchTipProperties(
            enable=False,
            params=LiquidClassTouchTipParams(zOffset=0, mmFromEdge=1, speed=10),
        )
    )
    with pytest.raises(ValueError):
        tp.z_offset = bad_value


@given(good_value=reasonable_numbers)
@settings(deadline=None, max_examples=50)
def test_touch_tip_properties_mm_from_edge(good_value: Union[int, float]) -> None:
    """Test valid mm_from_edge."""
    _build_touch_tip_properties(
        TouchTipProperties(
            enable=True,
            params=LiquidClassTouchTipParams(
                zOffset=0, mmFromEdge=good_value, speed=10
            ),
        )
    )
    tp = _build_touch_tip_properties(
        TouchTipProperties(
            enable=False,
            params=LiquidClassTouchTipParams(zOffset=0, mmFromEdge=1, speed=10),
        )
    )
    tp.mm_from_edge = good_value
    assert tp.mm_from_edge == float(good_value)


@given(bad_value=invalid_values)
@settings(deadline=None, max_examples=50)
def test_touch_tip_properties_mm_from_edge_bad_values(bad_value: Any) -> None:
    """Test invalid mm_from_edge values."""
    with pytest.raises(ValidationError):
        _build_touch_tip_properties(
            TouchTipProperties(
                enable=True,
                params=LiquidClassTouchTipParams(
                    zOffset=bad_value, mmFromEdge=1, speed=10
                ),
            )
        )
    tp = _build_touch_tip_properties(
        TouchTipProperties(
            enable=True,
            params=LiquidClassTouchTipParams(zOffset=0, mmFromEdge=1, speed=10),
        )
    )
    with pytest.raises(ValueError):
        tp.mm_from_edge = bad_value


@given(good_value=positive_non_zero_floats_and_ints)
@settings(deadline=None, max_examples=50)
def test_touch_tip_properties_speed(good_value: Union[int, float]) -> None:
    """Test valid speed."""
    _build_touch_tip_properties(
        TouchTipProperties(
            enable=True,
            params=LiquidClassTouchTipParams(zOffset=0, mmFromEdge=1, speed=good_value),
        )
    )
    tp = _build_touch_tip_properties(
        TouchTipProperties(
            enable=False,
            params=LiquidClassTouchTipParams(zOffset=0, mmFromEdge=1, speed=10),
        )
    )
    tp.speed = good_value
    assert tp.speed == float(good_value)


@given(bad_value=st.one_of(invalid_values, negative_or_zero_floats_and_ints))
@settings(deadline=None, max_examples=50)
def test_touch_tip_properties_speed_bad_values(bad_value: Any) -> None:
    """Test invalid speed values."""
    with pytest.raises(ValidationError):
        _build_touch_tip_properties(
            TouchTipProperties(
                enable=True,
                params=LiquidClassTouchTipParams(
                    zOffset=0, mmFromEdge=1, speed=bad_value
                ),
            )
        )
    tp = _build_touch_tip_properties(
        TouchTipProperties(
            enable=False,
            params=LiquidClassTouchTipParams(zOffset=0, mmFromEdge=1, speed=10),
        )
    )
    with pytest.raises(ValueError):
        tp.speed = bad_value



from typing import Protocol, Optional

from .types import VacuumModuleInfo, LEDColor, LEDPattern


class AbstractVacuumModuleDriver(Protocol):
    """Protocol for the Vacuum Module driver."""

    async def connect(self) -> None:
        """Connect to vacuum module."""
        ...

    async def disconnect(self) -> None:
        """Disconnect from vacuum module."""
        ...

    async def is_connected(self) -> bool:
        """Check connection to vacuum module."""
        ...

    async def get_device_info(self) -> VacuumModuleInfo:
        """Get Device Info."""
        ...

    async def set_serial_number(self, sn: str) -> None:
        """Set Serial Number."""
        ...

    async def enable_pump(self) -> None:
        """Enable the vacuum pump."""
        ...

    async def disable_pump(self) -> None:
        """Disable the vacuum pump."""
        ...

    async def get_pump_motor_register(self) -> None:
        """Get the register value of the pump motor driver."""
        ...

    async def get_pressure_sensor_register(self) -> None:
        """Get the register value of the pressure sensor driver."""
        ...

    async def get_pressure_sensor_reading_psi(self) -> float:
        """Get a reading from the pressure sensor."""
        ...

    async def set_vacuum_chamber_pressure(
        self,
        gage_pressure_mbarg: float,
        duration: Optional[float],
        rate: Optional[float],
    ) -> None:
        """Engage or release the vacuum until a desired internal pressure is reached."""
        ...

    async def get_gage_pressure_reading_mbarg(self) -> float:
        """Read each pressure sensor and return the pressure difference."""
        return 0.0

    # TODO: change pump power to be more specific when we find out how were gonna operate that
    async def engage_vacuum(self, pump_power: Optional[float] = None) -> None:
        """Engage the vacuum without regard to chamber pressure."""
        ...

    async def disengage_vacuum_pump(self) -> None:
        """Stops the vacuum pump, doesn't vent air or disable the motor."""
        ...

    async def vent(self) -> None:
        """Release the vacuum in the module chamber."""
        ...

    async def set_led(
        self,
        power: float,
        color: Optional[LEDColor] = None,
        external: Optional[bool] = None,
        pattern: Optional[LEDPattern] = None,
        duration: Optional[int] = None,  # Default firmware duration is 500ms
        reps: Optional[int] = None,  # Default firmware reps is 0
    ) -> None:
        """Set LED Status bar color and pattern."""
        ...

    async def enter_programming_mode(self) -> None:
        """Reboot into programming mode"""
        ...

    async def reset_serial_buffers(self) -> None:
        """Reset the input and output serial buffers."""
        ...



"""Tests for delay properties in the Opentrons protocol API."""

from pydantic import ValidationError
import pytest
from typing import Any, Union
from hypothesis import given, strategies as st, settings

from opentrons.protocol_api._liquid_properties import _build_delay_properties
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    DelayProperties,
    DelayParams,
)

from . import (
    boolean_looking_values,
    invalid_values,
    negative_non_zero_floats_and_ints,
    positive_or_zero_floats_or_ints,
)


def test_delay_properties_enable_and_disable() -> None:
    """Test enabling and disabling for boolean-only delay properties."""
    dp = _build_delay_properties(
        DelayProperties(enable=False, params=DelayParams(duration=1))
    )
    dp.enabled = True
    assert dp.enabled is True
    dp.enabled = False
    assert dp.enabled is False


def test_delay_properties_none_instantiation_combos() -> None:
    """Test handling of None combinations in delay properties instantiation."""
    with pytest.raises(ValidationError):
        _build_delay_properties(
            DelayProperties(enable=None, params=DelayParams(duration=None))  # type: ignore
        )
    with pytest.raises(ValidationError):
        _build_delay_properties(
            DelayProperties(enable=False, params=DelayParams(duration=None))  # type: ignore
        )
    with pytest.raises(ValidationError):
        _build_delay_properties(
            DelayProperties(enable=None, params=DelayParams(duration=1))  # type: ignore
        )
    with pytest.raises(ValidationError):
        _build_delay_properties(
            DelayProperties(enable=True, params=DelayParams(duration=None))  # type: ignore
        )


@given(bad_value=st.one_of(invalid_values, boolean_looking_values))
@settings(deadline=None, max_examples=50)
def test_delay_properties_enabled_bad_values(bad_value: Any) -> None:
    """Test bad values for DelayProperties.enabled."""
    with pytest.raises(ValidationError):
        _build_delay_properties(
            DelayProperties(enable=bad_value, params=DelayParams(duration=1))
        )
    dp = _build_delay_properties(
        DelayProperties(enable=True, params=DelayParams(duration=1))
    )
    with pytest.raises(ValueError):
        dp.enabled = bad_value


@given(good_duration=positive_or_zero_floats_or_ints)
@settings(deadline=None, max_examples=50)
def test_delay_properties_duration(good_duration: Union[int, float]) -> None:
    """Test valid float/int >= 0 for DelayProperties."""
    _build_delay_properties(
        DelayProperties(enable=False, params=DelayParams(duration=good_duration))
    )
    dp = _build_delay_properties(
        DelayProperties(enable=True, params=DelayParams(duration=1))
    )
    dp.duration = good_duration
    assert dp.duration == float(good_duration)


@given(bad_duration=st.one_of(negative_non_zero_floats_and_ints, invalid_values))
@settings(deadline=None, max_examples=50)
def test_delay_properties_duration_bad_values(bad_duration: Any) -> None:
    """Test invalid float/int for DelayProperties (must be >= 0)."""
    # instantiation
    with pytest.raises(ValidationError):
        _build_delay_properties(
            DelayProperties(enable=True, params=DelayParams(duration=bad_duration))
        )
    with pytest.raises(ValidationError):
        _build_delay_properties(
            DelayProperties(enable=False, params=DelayParams(duration=bad_duration))
        )
    # setting
    dp = _build_delay_properties(
        DelayProperties(enable=True, params=DelayParams(duration=1))
    )
    with pytest.raises(ValueError):
        dp.duration = bad_duration
    dp = _build_delay_properties(
        DelayProperties(enable=False, params=DelayParams(duration=1))
    )
    with pytest.raises(ValueError):
        dp.duration = bad_duration



"""Tests for mix properties in the Opentrons protocol API."""

from pydantic import ValidationError
import pytest
from typing import Any, Union
from hypothesis import given, strategies as st, settings

from opentrons.protocol_api._liquid_properties import _build_mix_properties
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    MixProperties,
    MixParams,
)

from . import (
    boolean_looking_values,
    invalid_values,
    positive_non_zero_floats_and_ints,
    negative_or_zero_floats_and_ints,
)


def test_mix_properties_enable_and_disable() -> None:
    """Test enabling and disabling for boolean-only mix properties."""
    mp = _build_mix_properties(
        MixProperties(enable=False, params=MixParams(repetitions=2, volume=10))
    )
    mp.enabled = True
    assert mp.enabled is True
    mp.enabled = False
    assert mp.enabled is False


def test_mix_properties_none_instantiation_combos() -> None:
    """Test handling of None combos in mix properties instantiation."""
    with pytest.raises(ValidationError):
        _build_mix_properties(MixProperties(enable=True, params=None))
    with pytest.raises(ValidationError):
        _build_mix_properties(MixProperties(enable=None, params=MixParams(repetitions=2, volume=10)))  # type: ignore
    _build_mix_properties(MixProperties(enable=False, params=None))


@given(bad_value=st.one_of(invalid_values, boolean_looking_values))
@settings(deadline=None, max_examples=50)
def test_mix_properties_enabled_bad_values(bad_value: Any) -> None:
    """Test bad values for MixProperties.enabled."""
    with pytest.raises(ValidationError):
        _build_mix_properties(
            MixProperties(enable=bad_value, params=MixParams(repetitions=1, volume=5))
        )
    mp = _build_mix_properties(
        MixProperties(enable=True, params=MixParams(repetitions=2, volume=10))
    )
    with pytest.raises(ValueError):
        mp.enabled = bad_value


@given(good_volume=positive_non_zero_floats_and_ints)
@settings(deadline=None, max_examples=50)
def test_mix_properties_volume_good_values(good_volume: Union[int, float]) -> None:
    """Test valid float/int > 0 for MixProperties volume."""
    mp = _build_mix_properties(
        MixProperties(enable=True, params=MixParams(repetitions=2, volume=5))
    )
    mp.volume = good_volume
    assert mp.volume == float(good_volume)


@given(bad_volume=st.one_of(negative_or_zero_floats_and_ints, invalid_values))
@settings(deadline=None, max_examples=50)
def test_mix_properties_volume_bad_values(bad_volume: Any) -> None:
    """Test invalid float/int <= 0 for MixProperties volume."""
    with pytest.raises(ValidationError):
        _build_mix_properties(
            MixProperties(
                enable=True, params=MixParams(repetitions=2, volume=bad_volume)
            )
        )
    mp = _build_mix_properties(
        MixProperties(enable=True, params=MixParams(repetitions=2, volume=5))
    )
    with pytest.raises(ValueError):
        mp.volume = bad_volume


@given(good_reps=st.integers(min_value=0, max_value=100))
@settings(deadline=None, max_examples=50)
def test_mix_properties_repetitions_good_values(good_reps: int) -> None:
    """Test valid int >= 0 for MixProperties repetitions."""
    _build_mix_properties(
        MixProperties(enable=True, params=MixParams(repetitions=good_reps, volume=5))
    )
    mp = _build_mix_properties(
        MixProperties(enable=True, params=MixParams(repetitions=2, volume=5))
    )
    mp.repetitions = good_reps
    assert mp.repetitions == good_reps


@given(bad_reps=st.one_of(st.integers(max_value=-1), invalid_values))
@settings(deadline=None, max_examples=50)
def test_mix_properties_repetitions_bad_values(bad_reps: Any) -> None:
    """Test invalid repetitions < 1 or non-integer."""
    with pytest.raises(ValidationError):
        _build_mix_properties(
            MixProperties(enable=True, params=MixParams(repetitions=bad_reps, volume=5))
        )
    mp = _build_mix_properties(
        MixProperties(enable=True, params=MixParams(repetitions=3, volume=5))
    )
    with pytest.raises(ValueError):
        mp.repetitions = bad_reps



from .driver import VacuumModuleDriver
from .simulator import SimulatingDriver
from .abstract import AbstractVacuumModuleDriver

__all__ = ["VacuumModuleDriver", "SimulatingDriver", "AbstractVacuumModuleDriver"]



"""Camera related state and store resource.

Camera settings, particularly for enablement, can be quieried from via the Camera Provider callback.
However, here Camera settings may also be provided to override or supercede those provided by the callbacks.
"""
from dataclasses import dataclass
from typing import Optional
from opentrons.protocol_engine.actions import AddCameraSettingsAction
from opentrons.protocol_engine.resources.camera_provider import CameraSettings

from ._abstract_store import HasState, HandlesActions
from ..actions import Action


@dataclass
class CameraState:
    """State of Engine Camera override settings."""

    enablement_settings: Optional[CameraSettings]
    # todo(chb, 2025-10-28): Eventually we will want to extend this to include the camera configurations overrides (contrast, zoom, etc)


class CameraStore(HasState[CameraState], HandlesActions):
    """Camera container."""

    _state: CameraState

    def __init__(self) -> None:
        """Initialize a Camera store and its state."""
        self._state = CameraState(enablement_settings=None)

    def handle_action(self, action: Action) -> None:
        """Modify state in reaction to an action."""
        if isinstance(action, AddCameraSettingsAction):
            # Update the Camera Enablement settings to the newest override settings
            self._state.enablement_settings = action.enablement_settings


class CameraView:
    """Read-only engine created Camera state view."""

    _state: CameraState

    def __init__(self, state: CameraState) -> None:
        """Initialize the view of Camera state.

        Arguments:
            state: Camera dataclass used for tracking override settings for the camera.
        """
        self._state = state

    def get_enablement_settings(self) -> CameraSettings | None:
        """Get the enablement settings override currently in use. This will take priority over Camera Provider callback provided settings."""
        return self._state.enablement_settings
